/*
 * Ilist speedup code.
 * In the case
 *	find dir -xdev ...
 * we read the ilist for the filesystem mounted at `dir',
 * build a bitmap of inodes that can be ignored (i.e. that generate
 * no actions and whose pathnames aren't needed).
 * We put a hook in the code that slurps up a directory (savedir()),
 * so that ignore-able inodes are skipped at that point.
 *
 * If the number of inodes that generate no action w/o respect
 * to pathnames is large, this is a major win.
 * It loses for things like
 *	find dir -xdev -name ...
 *
 * This code only kicks in if -xdev is given, and the directory
 * is the root of a mount point.
 */
#include <config.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include "defs.h"

/* Bitmap support */
#define BITSPERBYTE 8
#define N (BITSPERBYTE*sizeof(int))
#define TESTBIT(s, i)	(s[i/N] &  (1 << (i%N)))
#define SETBIT(s, i)	 s[i/N] |=  1 << (i%N)
#define ALLOC_SET(n)	(int *) calloc((n+N)/N, sizeof(int))
static int *set;

extern int (*dirhook)();
static int nignore, ndir, nitest;
int ildebug;


/* return 0 if we should ignore this inum */
static int
hook(inum)
	int inum;
{
	return !TESTBIT(set, inum);
}

/* Ilist walker calls this for every inode */
static int
itest(il, st, inum)
	void *il;
	struct stat *st;
        int inum;
{

	/*
	 * We need all directories in order to build pathnames,
	 * so don't ignore them.
	 */
	nitest++;
	if (S_ISDIR(st->st_mode)) {
		ndir++;
		return 0;
	}
	/*
	 * Execute the predicate tree with pathname==NULL.
	 * If any predicate needs the pathname, or any
	 * action predicate is called, runcode() quits and returns false.
	 * Otherwise, we know that this inode won't generate any actions,
	 * so we can ignore it.
	 * Remember that by setting a bit in `set'.
	 */
	have_stat = true;
	if (runcode(NULL, st)) {
		SETBIT(set, inum);
		nignore++;
	}
	return 0;
}

/*
 * Upper-level code entry points.
 * If `path' is the root of a mounted disk,
 * allocate the bitmap, and walk the ilist.
 * If any inodes can be ignored, set `dirhook' for savedir().
 */
setilist(path)
	char *path;
{
	void *il, *ilist_open();

	if (getenv("ILDEBUG") != NULL)
		ildebug++;
	if ((il = ilist_open(path)) == NULL)
		return;
	set = ALLOC_SET(ilist_nfiles(il));
	if (set != NULL) {
		ndir = nignore = 0;
		if (ildebug) fprintf(stderr, "walking ilist\n");
		if (ilist_walk(il, itest) == 0 && nignore > 0)
			dirhook = hook;
		else {
			dirhook = NULL;
			free(set);
			set = NULL;
		}
		if (ildebug) fprintf(stderr, "%d inodes, %d candidates, %d dirs\n",
			nitest, nitest-(nignore+ndir), ndir);
	}
	ilist_close(il);
}

endilist()
{
	dirhook = NULL;
	if (set)
		free(set);
	if (ildebug) {
		extern int totaldirs, emptydirs;
		fprintf(stderr, "%d dirs, %d empty\n", totaldirs,emptydirs);
	}
}
